/* 
 * $RCSfile: machine.c,v $
 *
 * x-kernel v3.2
 *
 * Copyright (c) 1993,1991,1990,1996  Arizona Board of Regents
 *
 * $Log: machine.c,v $
 * Revision 1.4  1997/07/01 18:58:21  mjk
 * Make all threads PTHREAD_DETACHED
 *
 * Revision 1.3  1997/03/10 21:37:34  mjk
 * *** empty log message ***
 *
 * Revision 1.2  1996/03/11 18:02:23  mjk
 * Udptest passes w/ Linux 1.2.13 and Pthreads.  All signals are caught by
 * specific thread(s).
 *
 * Revision 1.1  1996/02/19  20:21:13  mjk
 * Initial revision
 *
 */

#include <stdio.h>
#include <errno.h>
#include <netdb.h>
#include <syslog.h>
#include <signal.h>
#include <sys/file.h>
#include <sys/time.h>
#include <sys/wait.h>
#include <sys/types.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <netinet/in.h>

#include "thread.h"
#include "xtype.h"
#include "xk_debug.h"
#include "machine.h"


int tracesystem=0;

struct int_vector
{
    SigHandler *handler;
    void *arg;
};

/* don't know how many sockets we should allow for */
/* update this in [udp,tcp].c too */
#define MAX_FDS 30
static struct int_vector ivec[MAX_FDS + 1];
static int ivec_in_use_mask;

void cancelSignalHandler(int sock)
{
    ivec_in_use_mask &= ~(1 << sock);
    xTrace2(system, TR_MAJOR_EVENTS,
	    "Cancelling handler for %d mask = %x", sock, ivec_in_use_mask);
}

void installSignalHandler(int sock, SigHandler f, void *arg)
{
    ivec[sock].handler = f;
    ivec[sock].arg = arg;
    ivec_in_use_mask |= (1 << sock);
    xTrace3(system, TR_MAJOR_EVENTS,
	    "Setting handler for %d, arg = %x, mask = %x",
	    sock, arg, ivec_in_use_mask);
}

static unsigned int heldmask = 0;

void xholdsignals(int sock)
{
    xTrace1(system, TR_EVENTS, "Holding file descriptor %d", sock);
    heldmask |= (1 << sock);
}

void xreleasesignals(int sock)
{
    xTrace1(system, TR_EVENTS, "Releasing file descriptor %d", sock);
    heldmask &= ~(1 << sock);
}

/*
 *  Interrupt Handler
 *
 *  This thread handles all input (each input is signalled by a SIGIO
 *  delivered to to the process).
 */

static void fixFileMask(void)
{
    int	mask, i;
    static struct timeval zero = { 0, 0};

    for (i = 0; i < MAX_FDS; i++)
    {
	if (ivec_in_use_mask & (1 << i))
	{
	    mask = (1 << i);
	    if ((select(32, (fd_set *) &mask, (fd_set *) 0,
			(fd_set *) 0, &zero) < 0) && (errno == EBADF))
	    {
		xTrace1(system, TR_MAJOR_EVENTS, "Turning off socket #%d", i);
		ivec_in_use_mask &= ~(1 << i);
	    }
	}
    }
}

static int do_select(int *fds)
{
    int n;
    struct timeval tyme = { 0, 0 };

    n = select(MAX_FDS, (fd_set *) fds, (fd_set *) NULL, (fd_set *) NULL,
	       &tyme);
    while (n <= 0)
    {
	if (n == -1)
	{
	    if (errno == EBADF)
	    {
		xTrace0(system, TR_MAJOR_EVENTS, "EBADF on select by "
			"input_handler");
		fixFileMask();
	    }
	    else if (errno != EINTR)  /*  Restart if interrupted */
	    {
		perror("select");
		return 0;
	    }
	}
	else if (n == 0)
	{
	    xTrace0(system, TR_MAJOR_EVENTS, "do_select: can't find any input");
	    return 0;
	}

        tyme.tv_sec = 0;
        tyme.tv_usec = 0;
        n = select(MAX_FDS, (fd_set *) fds, (fd_set *) NULL, (fd_set *) NULL,
		   &tyme);
    }
    return 1;
}

static void *input_handler(void *arg)
{
  int 	sig_caught;
  int 	self = (int) pthread_self();
  int 	cur_set;
  int 	fd_index;
  int		n;

  xTrace1(system, TR_EVENTS,
	  "Starting up input handler thread with id = %x", self);

  for (;;)
  {
    sigset_t		set;

	    /* The SIGIO signal has been blocked for ALL threads.
               Here we will synchronously wait for one to occur.  */
    
    xTrace0(system, TR_EVENTS, "Input handler waiting for signal");
    sigemptyset(&set);
    sigaddset(&set, SIGIO);
    
	    /* WARNING: This version of sigwait always returns true.
	       It would be nicer it it followed Solaris and returned
	       the number of pending signals.  Since it does not the
	       xkernel WILL loose signals at high rates.  When this
	       happens the xkernel will block even thought data is
	       available for reading.  The next SIGIO will wakeup this
	       thread and both data batches will be read.  As long as
	       things keep moving you may never notice this.  But if
	       a high rate suddenly stops you may miss the tail end of
	       the data.  If the POSIX thread library would count the
	       pending signals and return from sigwait once for every
	       pending signal things would work.  Unfortunately the
	       library does not do this.  The alternative is the fix
	       the library (does POSIX specify this behavior?) or try
	       looping on a timed select. */

    (void)sigwait(&set, &sig_caught);
    xTrace0(system, TR_EVENTS, "Input handler received signal");
	
    cur_set = ivec_in_use_mask & ~heldmask;
    n = do_select(&cur_set);

    while (n && cur_set)
    {
      for (fd_index = 0; fd_index < MAX_FDS; fd_index++)
      {
	if (cur_set & (1 << fd_index))
	{
	  if (!ivec[fd_index].handler)
	  {
	    printf("Null handler for sock %d\n", fd_index);
	  }
	  else
	  {
	    xTrace1(system, TR_MAJOR_EVENTS,
		    "Calling handler for sock %d", fd_index);
	    (*ivec[fd_index].handler)(ivec[fd_index].arg);
	  }
	}
      }
      cur_set = ivec_in_use_mask & ~heldmask;
      n = do_select(&cur_set);
    }
  }
  return NULL;
}

void init_input_handler(void)
{
  static pthread_attr_t	io_attribs;
  pthread_t		io_thr;
  struct sched_param	io_sched;

				/* Create the I/O signal catcher
				   thread */
  pthread_attr_init(&io_attribs);
  pthread_attr_setdetachstate(&io_attribs, 1);
  pthread_attr_getschedparam(&io_attribs, &io_sched);
  io_sched.sched_priority = 100; /* NOT POSIX 				*/
  pthread_attr_setschedparam(&io_attribs, &io_sched);

  pthread_create(&io_thr, &io_attribs, input_handler, NULL);
}

