/*
 * $RCSfile: cookie.c,v $
 *
 * x-kernel v3.2
 *
 * $Log: cookie.c,v $
 * Revision 1.3  1996/02/15 21:57:55  mjk
 * added conditional #undef min before #define
 *
 */

/*
 *  The cookie protocol implements an idea proposed by Phil Karn
 *  to lessen the potential for "clogging" attacks; it enforces
 *  the provision that a remote participant must present a unique number
 *  assigned to it by the local host before communication can proceed.
 *
 *  The unique number is computed by using a secret value assigned
 *  at startup and the remote host address as inputs to a one-way hash
 *  function.  The number is sent to the remote address on request.
 *  An intruder must fake the remote address and see the response in
 *  order to get enough information to get past the cookie filter.
 *
 *  This is used in conjunction with computationally expensive cryptographic
 *  algorithms.
 *
 *  There are two message types: request and response.
 *  The request has no arguments and a response is always sent.
 *  The response has two elements: the local and remote cookies.
 *    If the local cookie matches, the remote cookie is recorded
 *    and the message is allowed
 *    If the local cookie does not match, the message is not accepted
 *    and no information from it is used.
 *  If a message is push'd to a host for which we have no cookie,
 *    the message is held while a cookie request message is sent.
 *    Any valid incoming cookie for the remote host will release
 *    the waiting thread.
 *
 */

int tracecookiep;

#include "xkernel.h"
#include "md5hash.h"
#include "cookie_i.h"

#ifdef __STDC__

void			cookie_init( XObj );
static XObj		cookieOpen( XObj, XObj, XObj, Part * );
static xkern_return_t	cookieOpenDone( XObj, XObj, XObj, XObj );
/* static xkern_return_t	cookieOpenDisable( XObj, XObj, XObj, Part * );*/
static xkern_return_t	cookieOpenEnable( XObj, XObj, XObj, Part * );
static xmsg_handle_t	cookiePush( XObj, Msg * );
static xkern_return_t	cookiePop( XObj, XObj, Msg *, void * );
static xkern_return_t	cookieDemux( XObj, XObj, Msg * );
static int		cookieControlProtl( XObj, int, char *, int ) ;
static int		cookieControlSessn( XObj, int, char *, int ) ;
static int		xCopyParts(Part **, char **, Part *, XObj);
static void		cookieCreateSessn( XObj );
static unsigned int	lincon1(unsigned int);
static unsigned int	lincon2(unsigned int);
static unsigned int	lincon3(unsigned int);
static cookie_t *	new_cookie(HOSTTYPE *, XObj) ;
static void		cookiePut(void *, char *, long, void *);
static void		cookieMsgPut(void *, char *, long, void *);
static long		cookieTake(void *, char *, long, void *);
static void		get_cookie(Event, void *);
static void		send_cookie_response(XObj);
static void		newdough(PState *);
static void		creq(void *, char *, long, void *);
#else
static XObj		cookieOpen( );
static xkern_return_t	cookieOpenDone( );
/* static xkern_return_t	cookieOpenDisable( ); */
static xkern_return_t	cookieOpenEnable( );
static xmsg_handle_t	cookiePush( );
static xkern_return_t	cookiePop( );
static xkern_return_t	cookieDemux( );
static int		xCopyParts( );
static void		cookieCreateSessn( );
static unsigned int	lincon1();
static unsigned int	lincon2();
static unsigned int	lincon3();
static cookie_t *	new_cookie();
static long		cookiePut();
static long		cookieTake();
static void		get_cookie();
static void		send_cookie_response();
static void		newdough();
static void		creq();
#endif

#define xGetParticipants(lls, p, l) \
  (xControl(lls, GETPARTICIPANTS, (char *)p, l))

static int
xCopyParts(q, qbuf, p, lls)
     Part **q;
     char **qbuf;
     Part *p;
     XObj lls;
{
  char *ebuf;
  int	len = 500;

  ebuf = xMalloc(500);
  if (lls != ERR_XOBJ) {
    if (xGetParticipants(lls, ebuf, 500)<0) return 0;
  }
  else
    partExternalize(p, ebuf, &len);
  *q = (Part *)xMalloc(partExtLen(ebuf)*sizeof(Part));
  partInternalize(*q, ebuf);
  *qbuf = ebuf;
  return partLen(((Part *)(*q)));
}

static XObj cookieOpen(self, hlp, hlpType, part)
     XObj self, hlp, hlpType;
     Part *part;
{
  XObj lls, sself = ERR_XOBJ;
  PState *ps = (PState *)(self->state);
  Part *partrem;
  HOSTTYPE *host;
  int len;
  char *ebuf = 0;

  len = xCopyParts(&partrem, &ebuf, part, ERR_XOBJ);
  if (len >= REMOTE_PART) host = (HOSTTYPE *)partPop(partrem[REMOTE_PART]);
  if (mapResolve(ps->cookieMap, host, &sself) == XK_SUCCESS) {
    SState *ss = (SState *)sself->state;

    if (!ss->this_cookie)
      ss->this_cookie = new_cookie(host, self);
    if (ebuf) { xFree(ebuf); xFree((char *)partrem); }
  }

  if ((lls = xOpen(self, hlpType, xGetDown(self, 0), part)) != ERR_XOBJ)
    {
      SState *ss;

      if (sself == ERR_XOBJ) {
	Msg m;

	sself = xCreateSessn(cookieCreateSessn, hlpType, hlpType, self, 1, &lls);
	mapBind(ps->cookieMap, host, sself);
	ss  = (SState *)sself->state;
	ss->this_cookie = new_cookie(host, self);
	ss->remhost = *host;
	msgConstructEmpty(&m);
	msgPush(&m, creq, 0, COOKIE_NET_TYPE_SIZE+HASH_SIZE_BYTES, ss);
	ss->waiting = 1;
	xPush(lls, &m);
	semWait(&ss->cookie_wait);
	ss->waiting = 0;
	xTrace0(cookiep, TR_EVENTS, "cookieOpen wakes");
      }
      if (ebuf) { xFree(ebuf); xFree((char *)partrem); }
      return sself;
    }
  if (ebuf) { xFree(ebuf); xFree((char *)partrem); }
  return ERR_XOBJ;
}

static void
cookieCreateSessn( self )
    XObj	self;
{
  SState *ss;

  xTrace1(cookiep, TR_EVENTS, "cookieCreateSessn entered: args %x", self);

  ss = X_NEW(SState);
  self->state = ss;
  semInit(&ss->cookie_wait, 0);
  self->push = cookiePush;
  self->pop = cookiePop;
  self->control = cookieControlSessn;
}

static xkern_return_t cookieOpenEnable(self, hlp, hlpType, p)
     XObj self, hlp, hlpType;
     Part *p;
{
  /* we might have the cookie, or it might be arriving */
  /* no need to create any state yet; wait for openDone */
  mapBind(((PState *)(self->state))->cookiePassMap, hlpType, hlp);
  return xOpenEnable(self, hlpType, xGetDown(self, 0), p);
}

static xkern_return_t	cookieOpenDone( self, lls, llp, hlpType  )
     XObj self, lls, llp, hlpType;
{
  Part *parts;
  HOSTTYPE *host;
  PState *ps = (PState *)self->state;
  XObj sessn, hlp;
  char *ebuf = 0;
  xkern_return_t ret = XK_FAILURE;

  xCopyParts(&parts, &ebuf, 0, lls);
  host = partPop(parts[REMOTE_PART]);
  if (ret = mapResolve(ps->cookiePassMap, hlpType, &hlp)==XK_SUCCESS) {
    if (mapResolve(ps->cookieMap, host, &sessn) != XK_SUCCESS)  {
      sessn = xCreateSessn(cookieCreateSessn, hlp, hlpType, self,
			   1, &lls);
      mapBind(ps->cookieMap, host, sessn);
      if (ebuf) { xFree((char *)parts); xFree(ebuf); ebuf = 0; }
      if (sessn == ERR_XOBJ) return XK_FAILURE;
    }
    xDuplicate(lls);
  }
  if (ebuf) { xFree((char *)parts); xFree(ebuf); }
      return xOpenDone(hlp, sessn, self);
  return ret;
}

static cookie_t null_cookie;

/*
 *
 *   cookiePush
 *
 *
 */
static xmsg_handle_t cookiePush(self, msg)
     XObj self;
     Msg *msg;
{
  SState *ss = (SState *)self->state;

  if (!ss->this_cookie)
    ss->this_cookie = new_cookie(&ss->remhost, self);

  if (!ss->that_cookie)
    /* don't know remote cookie; request it and wait */
    while (!ss->that_cookie) {
      semInit(&ss->cookie_wait, 0);
      evSchedule(get_cookie, ss, COOKIE_WAIT);
    }
  msgPush(msg, cookieMsgPut, 0, sizeof(cookie_msg_t), ss);

  return xPush(xGetDown(self, 0), msg);
}

/*
 *  creq  --- the cookie request processing function
 *
 */
static void creq(src, dst, len, ss)
    void *src;
    char *dst;
    long len;
    void *ss;
{
  SState *sf = (SState *)ss;

  *dst++ = COOKIE_REQUEST;
  bcopy((char *)sf->this_cookie, dst, HASH_SIZE_BYTES);
}

/*
 *  get_cookie
 *
 *  runs when wait is over
 *      what does this do?
 *
 */
static void get_cookie(ev, arg)
     Event ev;
     void *arg;
{
  /* COOKIE_REQUST */
  Msg msg;
  cookie_t cookie;
  XObj self = (XObj)arg;

  msgConstructEmpty(&msg);
  msgPush(&msg, creq, &cookie, sizeof(cookie_msg_t), 0);
  xPush(xGetDown(self, 0), &msg);
}

static void cookiePut(src, dst, len, sst)
    void *src;
    char *dst;
    long len;
    void *sst;
{
  bcopy(src, dst, HASH_SIZE_BYTES);
}

static void cookieMsgPut(src, dst, len, sst)
    void *src;
    char *dst;
    long len;
    void *sst;
{
  SState *ss = (SState *)sst;

  *dst = COOKIE_RESPONSE;
  dst += COOKIE_NET_TYPE_SIZE;
  if (ss->this_cookie)
    bcopy((char *)ss->this_cookie, dst, HASH_SIZE_BYTES);
  else
    xTrace0(cookiep, TR_ERRORS, "cookie: no local cookie");
  dst += HASH_SIZE_BYTES;
  
  if (ss->that_cookie)
    bcopy((char *)ss->that_cookie, dst, HASH_SIZE_BYTES);
  else
    xTrace0(cookiep, TR_ERRORS, "cookie: no remote cookie");
}

static long cookieTake(dst, src, len, m)
    void *dst;
    char *src;
    long len;
    void *m;
{
  cookie_msg_t *cm = (cookie_msg_t *)dst;

  cm->type = *src;
  src += COOKIE_NET_TYPE_SIZE;
  len -= COOKIE_NET_TYPE_SIZE;
  if (len < HASH_SIZE_BYTES) return COOKIE_NET_TYPE_SIZE;
  bcopy(src, cm->this_cookie, HASH_SIZE_BYTES);
  src += HASH_SIZE_BYTES;
  len -= HASH_SIZE_BYTES;
  if (len < HASH_SIZE_BYTES) return COOKIE_NET_TYPE_SIZE;
  bcopy(src, cm->that_cookie, HASH_SIZE_BYTES);
  return COOKIE_NET_TYPE_SIZE + 2 * HASH_SIZE_BYTES;
}

static void send_cookie_response(self)
     XObj self;
{
  Msg msg;

  msgConstructEmpty(&msg);
  msgPush(&msg, cookieMsgPut, 0, 0, (void *)self->state);
  xPush(xGetDown(self, 0), &msg);
  msgDestroy(&msg);
}

#ifdef min
#undef min
#endif
#define min(a, b) ( ((a)<=(b))?a:b )

/*
 *  cookieDemux
 *
 *   1. ordinary case requires matching incoming cookie, may update remote
 *   2. requests for cookies require nothing, always get a response
 *   3. responses require nothing, always update remote
 *
 */
static xkern_return_t cookieDemux(self, lls, msg)
     XObj self, lls;
     Msg  *msg;
{
  Part *part;
  PState *ps = (PState *)self->state;
  XObj sself;
  HOSTTYPE *remhost;
  cookie_msg_t cookieIn;
  char *ebuf ;
  int mlen;

  mlen = msgLen(msg);
  msgPop(msg, cookieTake, &cookieIn, min(mlen, sizeof(cookie_msg_t)), 0);
  xCopyParts(&part, &ebuf, 0, lls);
  remhost = (IPhost *)partPop(part[REMOTE_PART]);
  if (mapResolve(ps->cookieMap, remhost, &sself) == XK_SUCCESS) {
    SState *ss = (SState *) sself->state;

    /* if remote side wants a cookie, make one if necessary and send it */
    if (cookieIn.type == COOKIE_REQUEST) {
      if (!ss->this_cookie)
	ss->this_cookie = new_cookie(remhost, self);
      if (!ss->that_cookie) ss->that_cookie = X_NEW(cookie_t);
      bcopy((char *)&cookieIn.this_cookie, 
	    (char *)ss->that_cookie, sizeof(cookie_t)); 
      msgConstructEmpty(&ss->msg);
      msgPush(&ss->msg, cookieMsgPut, 0, sizeof(cookie_msg_t), ss);
      xPush(lls, &ss->msg);
    }
      /* if remote side has the right cookie for me and offers a new,
	 non-trivial cookie, eat it */
    else
      if (ss->that_cookie &&
	  !bcmp ((char *)&cookieIn.that_cookie, (char *)ss->this_cookie, HASH_SIZE_BYTES) &&
	   bcmp ((char *)&cookieIn.this_cookie, (char *)&null_cookie, HASH_SIZE_BYTES) &&
	   bcmp ((char *)&cookieIn.this_cookie, (char *)ss->that_cookie, HASH_SIZE_BYTES))
	{
	  bcopy((char *)&cookieIn.this_cookie, (char *)ss->that_cookie, sizeof(cookie_t));
	  if (cookieIn.type == COOKIE_RESPONSE && ss->waiting)
	    semSignal(&ss->cookie_wait);
	}
      else if ( !ss->that_cookie && cookieIn.type == COOKIE_RESPONSE &&
	       !bcmp ((char *)&cookieIn.that_cookie, (char *)ss->this_cookie,
		     HASH_SIZE_BYTES))
	{
	  /* he's got our cookie, now we have his */
	  if (!ss->that_cookie)  ss->that_cookie = X_NEW(cookie_t);
	  bcopy((char *)&cookieIn.this_cookie, (char *)ss->that_cookie,
		sizeof(cookie_t));
	  semSignal(&ss->cookie_wait);
	}
    /* otherwise, everything must match */
      else if (
	       bcmp (cookieIn.that_cookie, (char *)ss->this_cookie, HASH_SIZE_BYTES)  ||
	       !bcmp (cookieIn.this_cookie, (char *)&null_cookie, HASH_SIZE_BYTES) ||
	       bcmp (cookieIn.this_cookie, (char *)ss->that_cookie, HASH_SIZE_BYTES))
	{
	  if (ebuf) { xFree((char *)part); xFree(ebuf); }
	  return XK_FAILURE;
	}
  }
  if (ebuf) { xFree((char *)part); xFree(ebuf);}
  if (msgLen(msg)>0)
    return xPop(sself, lls, msg, 0);
  else return XK_SUCCESS;
}


static  xkern_return_t
cookiePop ( self, lls, msg, hdr)
     XObj self, lls;
     Msg *msg;
     void *hdr;
{
  return xDemux(self, msg);
}

static int
cookieControlSessn(s, opcode, buf, len)
    XObj s;
    int opcode;
    char *buf;
    int len;
{
    xTrace0(cookiep, TR_FULL_TRACE, "cookie control session");
    return xControl(xGetDown(s, 0), opcode, buf, len);;
}

static int
cookieControlProtl(s, opcode, buf, len)
    XObj s;
    int opcode;
    char *buf;
    int len;
{
    xTrace0(cookiep, TR_FULL_TRACE, "cookie control protocol");
    return xControl(xGetDown(s, 0), opcode, buf, len);;
}


static unsigned int lincon1(p) 
    unsigned  int p;
{
  static unsigned int state = 1;

  state  = (state * 2) % p;
  return state;
}

static unsigned int lincon2(p)
     unsigned int p;
 {
  static unsigned int state = 2;

  state = (state * 3) % p;
  return state;
}

static unsigned int lincon3(p)
     unsigned int p;
 {
  static unsigned int state = 3;

  state = (state * 5) % p;
  return state;
}


/*
 *  allocate and compute a cookie unique to the remote host
 *
 *
 */
static cookie_t *new_cookie(host, self)
  HOSTTYPE *host;
  XObj	self;
{
  MD5_CTX md5ctx;
  cookie_t *cookie = X_NEW(cookie_t);
  PState *ps = (PState *)self->state;

  MD5Init(&md5ctx);
  MD5Update(&md5ctx, (unsigned char *)&ps->bowl, (unsigned int) sizeof(ps->bowl));
  MD5Update(&md5ctx, (char *)host, sizeof(host));
  MD5Final(&md5ctx);
  bcopy(md5ctx.digest, (char *)cookie, HASH_SIZE_BYTES);
  return cookie;
}

/*
 *  get a new starting value for creating cookies
 *
 */
static void newdough(ps)
     PState *ps;
{
  XTime xtime;

  xGetTime(&xtime);

  ps->bowl[0] = ps->bowl[3] ^ xtime.sec 
    ^ lincon1(ps->p1) ^ lincon2(ps->p2) ^ lincon3(ps->p3);
  ps->bowl[1] = ps->bowl[2] ^ xtime.usec 
    ^ lincon1(ps->p1) ^ lincon2(ps->p2) ^ lincon3(ps->p3);
  ps->bowl[2] = ps->bowl[1] ^ xtime.usec 
    ^ lincon1(ps->p1) ^ lincon2(ps->p2) ^ lincon3(ps->p3);
  ps->bowl[3] = ps->bowl[0] ^ xtime.sec 
    ^ lincon1(ps->p1) ^ lincon2(ps->p2) ^ lincon3(ps->p3);
}

static xkern_return_t romranseed(XObj, char **, int, int, void *);

static XObjRomOpt rmopts[1] = {{ "randomseed", 1, romranseed }};

static xkern_return_t
romranseed(self, ptr, fields, line, arg)
XObj self;
char **ptr;
int  fields, line;
void *arg;
{
  int n;
  unsigned int m;
  PState *ps = (PState *)(self->state);

  if (atoi(ptr[2]) > 0 && (n = atoi(ptr[2]) < 4)) {
    m = atoi(ptr[3]);
    if (m != 0) {
      switch (n) {
      case 1: ps->p1 = m; break;
      case 2: ps->p2 = m; break;
      case 3: ps->p3 = m; break;
      }
      return XK_SUCCESS;
    }
    else xTrace0(cookiep, TR_SOFT_ERROR, "cookie rom init value zero rejected");
  }
  else xTrace1(cookiep, TR_SOFT_ERROR, "cookie rom index value invalid %s", ptr[2]);
  return XK_FAILURE;
}

void cookie_init( self )
     XObj self;
{
  PState *ps;

  xTrace0(cookiep, TR_GROSS_EVENTS, "cookie initializes");

  self->state = ps = X_NEW(PState);
  ps->p1 = 2000000011;
  ps->p2 = 2000000033;
  ps->p3 = 2000000063;
  newdough(ps);
  ps->cookieMap = mapCreate(COOKIE_MAP_SIZE, sizeof(HOSTTYPE));
  ps->cookiePassMap = mapCreate(COOKIE_MAP_SIZE, sizeof(XObj));
  self->open = cookieOpen;
  self->opendone = cookieOpenDone;
  self->openenable = cookieOpenEnable;
  self->pop = cookiePop;
  self->push = cookiePush;
  self->demux = cookieDemux;
  self->control = cookieControlProtl;
  findXObjRomOpts(self, rmopts, 1, 0);
}

